$ifndef _23K256_
$define _23K256_
' Interface to a Microchip 23K256 (32k byte) serial SRAM using the MSSP SPI peripheral
'
' Written by Les Johnson 31st May 2010
' For use with the Crownhill Amicus18 BASIC compiler version 1.0.0.8 onwards
'
' This source code is supplied as-is, and no guarantees are offered on it's operation
'
' User defines for the SPI pins
'
$ifndef SRAM_CS
    $define SRAM_CS PORTC.0             ' The 23K256 SRAM CS line is connected to this pin
$endif
$ifndef SRAM_CLK
    $define SRAM_CLK PORTC.3            ' Connects to the 23K256 CLK line
$endif
$ifndef SRAM_SDI
    $define SRAM_SDI PORTC.4            ' Connects to the 23K256 SDO line
$endif
$ifndef SRAM_SDO
    $define SRAM_SDO PORTC.5            ' Connects to the 23K256 SDI line
$endif
'
' Create variables and aliases for the routines
'
    Dim tSRAM_FromROM As EEADR.0            ' Indicates source from RAM or code memory
    Dim wSourceDestAddress As FSR1L.Word    ' Used to hold the address of the source or destination string
    Dim wSRAM_Address As PRODL.Word         ' Used to hold the address within the SRAM
    Dim wSRAM_Count As Word System          ' Used to count the amount of bytes read or written to the SRAM
    Dim bSRAM_Data As wSRAM_Count.Byte0     ' Holds the 8-bit value written to and read from the SRAM

    Dim wSRAM_FSR0 As FSR0L.Word            ' Create a 16-bit variable from SFR pair FSR0L\H
    Dim wSRAM_FSR1 As FSR1L.Word            ' Create a 16-bit variable from SFR pair FSR1L\H
    Dim wSRAM_Tblptr As TBLPTRL.Word        ' Create a 16-bit variable from SFR pair TBLPTRL\H
    Dim wSRAM_Prod As PRODL.Word            ' Create a 16-bit variable from SFR pair PRODL\H
'
' Define SPI commands for the SRAM
'
    $define cComRead $03                    ' Read from memory array beginning at the selected address
    $define cComWrite $02                   ' Write to memory array beginning at the selected address
    $define cComReadStatus $05              ' Read the status register
    $define cComWriteStatus $01             ' Write the status register

    $define cModeByte $01                   ' bit7 = 0 : bit6 = 0 - Byte Mode setting LSB 1 disables the hold functionality
    $define cModePage $81                   ' bit7 = 1 : bit6 = 0 - Page Mode
    $define cModeSeq $41                    ' bit7 = 0 : bit6 = 1 - Sequential mode

$ifndef True
    $define True 1
$endif
$ifndef False
    $define False 0
$endif
'-------------------------------------------------------------------------------------------------

    GoTo SRAM_Main                          ' Jump over the subroutines

'-------------------------------------------------------------------------------------------------
' Helper macro that checks the SPI interface BF flag
'
' Input   : None
' Output  : None
' Notes   : None
'
$define SRAM_WaitForReady() Btfss SSPSTATbits_BF : Bra $ - 2
'-------------------------------------------------------------------------------------------------
' Initialise the MSSP module for 4MHz SPI mode 1:1, CKP = 0, CKE = 1
'
' Input   : None
' Output  : None
' Notes   : None
'
$define SRAM_Init() _SRAM_Init

_SRAM_Init Macro-
    GoSub __SRAM_Init
Endm

#ifMacro- _SRAM_Init                        ' Add this block only if the macro has been used in the program
__SRAM_Init:
    Output SRAM_CS                          ' CS pin is output
    Output SRAM_CLK                         ' CLK pin is output
    Input SRAM_SDI                          ' SDI pin is input
    Output SRAM_SDO                         ' SDO pin is output
    Set SRAM_CS                             ' SRAM is disabled
    Clear SSPCON2                           ' Clear all the flags
    Clear SSPSTAT                           ' Input data sampled at the middle of data output time SMP = 0
    Set SSPSTATbits_CKE                     ' Select the SPI mode 1:1
    SSPCON1 = %00000001                     ' CKP = 0, SCK = FOSC/16 (4MHz operation)
    Set SSPCON1bits_SSPEN                   ' Enable the SCK, SDO, SDI pins
    Return
#endIfMacro-
'-------------------------------------------------------------------------------------------------
' Write a page to the SRAM
'
' Input     : FSR0L\H register pair hold the address of a RAM based source String
'           : TBLPTRL\H register pair hold the address of a code memory based source String
'           : wSRAM_Address holds the page to write within the SRAM
'           : Byte0 of wSRAM_Count holds the amount of bytes to write
' Output    : None
' Notes     : Configure SRAM for page mode before calling this routine
'           : User needs to make sure that they are writing within the page (32 bytes maximum)
'
$define SRAM_WritePage(pPage, pCount, pSource) '
    wSourceDestAddress = VarPtr pSource '
    wSRAM_Address = pPage '
    WREG = pCount '
    _SRAMWritePage pSource

_SRAMWritePage Macro- pSource
    #if(Prm_1 == Char)                      ' Is the pSource parameter a code memory string?
        tSRAM_FromROM = True                ' Yes. So signal this to the subroutine
    #else                                   ' Otherwise...
        tSRAM_FromROM = False               ' Signal a RAM based source
    #endif
    GoSub __SRAMWritePage
Endm

#ifMacro- _SRAMWritePage                    ' Add this block only if the macro has been used in the program
__SRAMWritePage:
    wSRAM_Count.Byte0 = WREG                ' Transfer the pCount parameter into its working variable
    If tSRAM_FromROM = True Then            ' Does the string reside in code memory?
        wSRAM_Tblptr = wSourceDestAddress   ' Yes. So transfer the address to register pair TBLPTRL\H
    Else                                    ' Otherwise...
        wSRAM_FSR0 = wSourceDestAddress     ' Transfer the address to register pair FSR0L\H
    EndIf
    wSRAM_Address = wSRAM_Address * 32      ' Locate the page
    Clear SRAM_CS                           ' Enable SRAM

    SSPBUF = cComWrite                      ' Write command to SRAM
    SRAM_WaitForReady()                     ' Check for the BF flag to ensure transmission
    WREG = SSPBUF                           ' Clear the BF flag

    SSPBUF = wSRAM_Address.Byte1            ' High Address byte of the SRAM
    SRAM_WaitForReady()                     ' Check for the BF flag to ensure transmission
    WREG = SSPBUF                           ' Clear the BF flag

    SSPBUF = wSRAM_Address.Byte0            ' Low Address byte of the SRAM
    SRAM_WaitForReady()                     ' Check for the BF flag to ensure transmission
    WREG = SSPBUF                           ' Clear the BF flag

    Repeat                                  ' Form a loop for the elements within the array
        Dec wSRAM_Count.Byte0               ' Count down the element iterations
        If tSRAM_FromROM = True Then        ' Does the string reside in code memory?
            Tblrd*+                         ' Yes. So read a byte from code memory and increment to the next address
            SSPBUF = TABLAT                 ' Write a byte to the SRAM from code memory
        Else                                ' Otherwise...
            SSPBUF = POSTINC0               ' Write a byte to the SRAM from RAM
        EndIf
        SRAM_WaitForReady()                 ' Check for the BF flag to ensure transmission
        WREG = SSPBUF                       ' Clear the BF flag
    Until wSRAM_Count.Byte0 = 0             ' Until wSRAM_Count elements are performed

    Set SRAM_CS                             ' Disable the SRAM
    Return
#endIfMacro-
'-------------------------------------------------------------------------------------------------
' Read a page from the SRAM
'
' Input     : wSRAM_Address holds the page to read from
'           : Byte0 of wSRAM_Count holds the amount of bytes to write
' Output    : FSR0L\H holds the address of the RAM that will be loaded with bytes read from the SRAM
' Notes     : Configure SRAM For page mode before calling this routine
'           : User needs To make sure that they are writing within the page (max 32 bytes)
'
$define SRAM_ReadPage(pPage,pCount, pDest) '
    wSRAM_FSR0 = VarPtr pDest '
    wSRAM_Address = pPage '
    WREG = pCount '
    _SRAMReadPage

_SRAMReadPage Macro-
    GoSub __SRAMReadPage
Endm

#ifMacro- _SRAMReadPage                     ' Add this block only if the macro has been used in the program
__SRAMReadPage:
    wSRAM_Count.Byte0 = WREG                ' Transfer the pCount parameter into its working variable
    wSRAM_Address = wSRAM_Address * 32      ' Locate the page
    Clear SRAM_CS                           ' Enable SRAM

    SSPBUF = cComRead                       ' Read command to SRAM
    SRAM_WaitForReady()                     ' Check for the BF flag to ensure transmission
    WREG = SSPBUF                           ' Clear the BF flag

    SSPBUF = wSRAM_Address.Byte1            ' High Address byte of the SRAM
    SRAM_WaitForReady()                     ' Check for the BF flag to ensure transmission
    WREG = SSPBUF                           ' Clear the BF flag

    SSPBUF = wSRAM_Address.Byte0            ' Low Address byte of the SRAM
    SRAM_WaitForReady()                     ' Check for the BF flag to ensure transmission
    WREG = SSPBUF                           ' Clear the BF flag

    Repeat                                  ' Form a loop for the elements within the array
        Dec wSRAM_Count.Byte0               ' Count down the element iterations
        SSPBUF = $FF                        ' Dummy write to the SSPBUF
        SRAM_WaitForReady()                 ' Check for the BF flag to verify data reception
        POSTINC0 = SSPBUF                   ' Read from SRAM to the destination RAM address
    Until wSRAM_Count.Byte0 = 0             ' Until wSRAM_Count elements are performed
    Set SRAM_CS                             ' Disable the SRAM
    Return
#endIfMacro-
'-------------------------------------------------------------------------------------------------
' Write sequentially to the SRAM
'
' Input     : FSR0L\H register pair hold the address of a RAM based source String
'           : TBLPTRL\H register pair hold the address of a code memory based source String
'           : wSRAM_Address holds the address to write too
'           : wSRAM_Count holds the amount of bytes to write
'           : Byte Array bSRAM_PageRAM holds the values to write
' Output    : None
' Notes     : In Sequence Mode we can write across the page boundary
'           : Make sure that the SRAM is configured for Sequential mode
'
$define SRAM_WriteSeq(pAddress, pCount, pSource) '
    wSourceDestAddress = VarPtr pSource '
    wSRAM_Address = pAddress '
    wSRAM_Count = pCount '
    _SRAMWriteSeq pSource

_SRAMWriteSeq Macro- pSource
    #if(Prm_1 == Char)                      ' Is the pSource parameter a code memory string?
        tSRAM_FromROM = True                ' Yes. So signal this to the subroutine
    #else                                   ' Otherwise...
        tSRAM_FromROM = False               ' Signal a RAM based source
    #endif
    GoSub __SRAMWriteSeq
Endm

#ifMacro- _SRAMWriteSeq                     ' Add this block only if the macro has been used in the program
__SRAMWriteSeq:
    If tSRAM_FromROM = True Then            ' Does the string reside in code memory?
        wSRAM_Tblptr = wSourceDestAddress   ' Yes. So transfer the address to register pair TBLPTRL\H
    Else                                    ' Otherwise...
        wSRAM_FSR0 = wSourceDestAddress     ' Transfer the address to register pair FSR0L\H
    EndIf
    Clear SRAM_CS                           ' Enable SRAM

    SSPBUF = cComWrite                      ' Write command to SRAM
    SRAM_WaitForReady()                     ' Check for the BF flag to ensure transmission
    WREG = SSPBUF                           ' Clear the BF flag

    SSPBUF = wSRAM_Address.Byte1            ' High Address byte of the SRAM
    SRAM_WaitForReady()                     ' Check for the BF flag to ensure transmission
    WREG = SSPBUF                           ' Clear the BF flag

    SSPBUF = wSRAM_Address.Byte0            ' Low Address byte of the SRAM
    SRAM_WaitForReady()                     ' Check for the BF flag to ensure transmission
    WREG = SSPBUF                           ' Clear the BF flag

    Repeat                                  ' Form a loop for the elements within the array
        Dec wSRAM_Count                     ' Count down the element iterations
        If tSRAM_FromROM = True Then        ' Does the source reside in code memory?
            Tblrd*+                         ' Yes. So read a byte from code memory and increment to the next address
            SSPBUF = TABLAT                 ' Write a byte to the SRAM from code memory
        Else                                ' Otherwise...
            SSPBUF = POSTINC0               ' Write a byte to the SRAM from RAM
        EndIf
        SRAM_WaitForReady()                 ' Check for the BF flag to ensure transmission
        WREG = SSPBUF                       ' Clear the BF flag
    Until wSRAM_Count = 0                   ' Until wSRAM_Count elements are performed
    Set SRAM_CS                             ' Disable the SRAM
    Return
#endIfMacro-

'-------------------------------------------------------------------------------------------------
' Read from SRAM in sequential mode
'
' Input     : wSRAM_Address holds the address to read from
'           : wSRAM_Count holds the amount of bytes to write
' Output    : FSR0L\H holds the address of the RAM that will be loaded with bytes read from the SRAM
' Notes     : Make sure that the SRAM is configured for Sequential mode
'
$define SRAM_ReadSeq(pAddress, pCount, pDest) '
    wSRAM_FSR0 = VarPtr pDest '
    wSRAM_Address = pAddress '
    wSRAM_Count = pCount '
    _SRAMReadSeq

_SRAMReadSeq Macro- pDest
    GoSub __SRAMReadSeq
Endm

#ifMacro- _SRAMReadSeq                      ' Add this block only if the macro has been used in the program
__SRAMReadSeq:
    Clear SRAM_CS                           ' Enable SRAM

    SSPBUF = cComRead                       ' Read command to SRAM
    SRAM_WaitForReady()                     ' Check for the BF flag to ensure transmission
    WREG = SSPBUF                           ' Clear the BF flag

    SSPBUF = wSRAM_Address.Byte1            ' High Address byte of the SRAM
    SRAM_WaitForReady()                     ' Check for the BF flag to ensure transmission
    WREG = SSPBUF                           ' Clear the BF flag

    SSPBUF = wSRAM_Address.Byte0            ' Low Address byte of the SRAM
    SRAM_WaitForReady()                     ' Check for the BF flag to ensure transmission
    WREG = SSPBUF                           ' Clear the BF flag

    Repeat                                  ' Form a loop for the elements within the array
        Dec wSRAM_Count                     ' Count down the element iterations
        SSPBUF = $FF                        ' Dummy write to the SSPBUF
        SRAM_WaitForReady()                 ' Check for the BF flag to verify bData reception
        POSTINC0 = SSPBUF                   ' Read from SRAM to the array
    Until wSRAM_Count = 0                   ' Until wSRAM_Count elements are performed

    Set SRAM_CS                             ' Disable the SRAM
    Return
#endIfMacro-

'-------------------------------------------------------------------------------------------------
' Write a Byte to SRAM
'
' Input     : wSRAM_Address holds the address to write too
'           : bSRAM_Data and WREG hold the value to write to SRAM
' Output    : None
' Notes     : Make sure that before calling this routine
'             the status register of SRAM is configured for Byte Mode
'
$define SRAM_WriteByte(pAddress, pValue) '
    wSRAM_Address = pAddress '
    WREG = pValue '
    _SRAMWriteByte

_SRAMWriteByte Macro-
    GoSub __SRAMWriteByte
Endm

#ifMacro- _SRAMWriteByte                    ' Add this block only if the macro has been used in the program
__SRAMWriteByte:
    bSRAM_Data = WREG                       ' Transfer the passed WREG to it's working variable
    Clear SRAM_CS                           ' Enable SRAM

    SSPBUF = cComWrite                      ' Write command to SRAM
    SRAM_WaitForReady()                     ' Check for the BF flag to ensure transmission
    WREG = SSPBUF                           ' Clear the BF flag

    SSPBUF = wSRAM_Address.Byte1            ' High Address byte of the SRAM
    SRAM_WaitForReady()                     ' Check for the BF flag to ensure transmission
    WREG = SSPBUF                           ' Clear the BF flag

    SSPBUF = wSRAM_Address.Byte0            ' Low Address byte of the SRAM
    SRAM_WaitForReady()                     ' Check for the BF flag to ensure transmission
    WREG = SSPBUF                           ' Clear the BF flag

    SSPBUF = bSRAM_Data                     ' Write to command to SRAM
    SRAM_WaitForReady()                     ' Check for the BF flag to ensure transmission
    WREG = SSPBUF                           ' Clear the BF flag

    Set SRAM_CS                             ' Disable the SRAM
    Return
#endIfMacro-

'-------------------------------------------------------------------------------------------------
' Read a Byte from SRAM
' Input     : wSRAM_Address holds the address to write too
' Output    : WREG holds the value read from the SRAM
' Notes     : Make sure that before calling this routine
'             the status register of SRAM is configured For Byte Mode
'
$define SRAM_ReadByte(pAddress) _SRAMReadByte pAddress

_SRAMReadByte Macro- pAddr
    #if(Prm_1 == Num8) || (Prm_1 == Num16) || (Prm_1 == Num32)
        num_word pAddr, wSRAM_Address
    #endif
    #if(Prm_1 == Byte)
        byte_word pAddr, wSRAM_Address
    #endif
    #if(Prm_1 == Word)
        word_word pAddr, wSRAM_Address
    #endif
    #if(Prm_1 == Dword)
        dword_word pAddr, wSRAM_Address
    #endif
    #if(Prm_1 == Float)
        #error "SRAM_ReadByte macro's address cannot be a floating point"
        Exitm
    #endif
    GoSub __SRAMReadByte
    Return_Wreg
Endm

#ifMacro- _SRAMReadByte                     ' Add this block only if the macro has been used in the program
__SRAMReadByte:
    Clear SRAM_CS                           ' Enable SRAM

    SSPBUF = cComRead                       ' Read command to SRAM
    SRAM_WaitForReady()                     ' Check for the BF flag to ensure transmission
    WREG = SSPBUF                           ' Clear the BF flag

    SSPBUF = wSRAM_Address.Byte1            ' High Address byte of the SRAM
    SRAM_WaitForReady()                     ' Check for the BF flag to ensure transmission
    WREG = SSPBUF                           ' Clear the BF flag

    SSPBUF = wSRAM_Address.Byte0            ' High Address byte of the SRAM
    SRAM_WaitForReady()                     ' Check for the BF flag to ensure transmission
    WREG = SSPBUF                           ' Clear the BF flag

    SSPBUF = $FF                            ' Dummy write to SSPBUF to read from SRAM
    SRAM_WaitForReady()                     ' Check for the BF flag to verify bData reception
    WREG = SSPBUF                           ' Copy the value read to the WREG

    Set SRAM_CS                             ' Disable the SRAM
    Return
#endIfMacro-
'-------------------------------------------------------------------------------------------------
' Change the status of the SRAM
'
' Input     : bSRAM_Data holds the status value
' Output    : None
' Notes     : None
'
$define SRAM_WriteStatus(pValue) '
    WREG = pValue '
    _SRAMWriteStatusReg

_SRAMWriteStatusReg Macro-
    GoSub __SRAMWriteStatusReg
Endm

#ifMacro- _SRAMWriteStatusReg               ' Add this block only if the macro has been used in the program
__SRAMWriteStatusReg:
    bSRAM_Data = WREG                       ' Transfer the passed WREG to it's working variable
    Clear SRAM_CS                           ' Enable SRAM

    SSPBUF = cComWriteStatus                ' Write Status Register command to SRAM
    SRAM_WaitForReady()                     ' Check for the BF flag to ensure transmission
    WREG = SSPBUF                           ' Clear the BF flag

    SSPBUF = bSRAM_Data                     ' Write the mode to SRAM
    SRAM_WaitForReady()                     ' Check for the BF flag to ensure transmission
    WREG = SSPBUF                           ' Clear the BF flag

    Set SRAM_CS                             ' Disable the SRAM
    Return
#endIfMacro-
'-------------------------------------------------------------------------------------------------
' Name    : SRAM_ReadStatusReg
' Purpose : Read the status of the SRAM
' Input   : None
' Output  : WREG holds the status value read from the SRAM
' Notes   : None
'
$define SRAM_ReadStatus() _SRAMReadStatusReg

_SRAMReadStatusReg Macro-
    GoSub __SRAMReadStatusReg
    Return_Wreg
Endm

#ifMacro- _SRAMReadStatusReg                ' Add this block only if the macro has been used in the program
__SRAMReadStatusReg:
    Clear SRAM_CS                           ' Enable SRAM

    SSPBUF = cComReadStatus                 ' Read Status Register command to SRAM
    SRAM_WaitForReady()                     ' Check for the BF flag to ensure transmission
    WREG = SSPBUF                           ' Clear the BF flag

    SSPBUF = $FF                            ' Dummy write to SSPBUF to read SRAM of SRAM
    SRAM_WaitForReady()                     ' Check for the BF flag to ensure transmission
    WREG = SSPBUF                           ' Copy content to value read to the WREG

    Set SRAM_CS                             ' Disable the SRAM
    Return
#endIfMacro-

'--------------------------------------------------------------------------------
SRAM_Main:
    SRAM_Init()                             ' Initalise the MSSP module for SPI master
$endif
